import sys
import os
import argparse
import logging
import aiohttp_cors
import socketio
import asyncio
import uuid

from aiohttp import web, streamer

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("socketio-server")
logger.setLevel(logging.INFO)


def create_socketio_server(static_dir=None):
    """Create and return aiohttp webserver app."""
    sio = socketio.AsyncServer(cors_allowed_origins="*")
    app = web.Application()
    sio.attach(app)
    setup_router(app, static_dir)
    setup_socketio(sio)
    setup_cors(app)
    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    return app


async def on_startup(app):
    """Run on server start."""
    logger.info("socketio-server started")


async def on_shutdown(app):
    """Run on server shut down."""
    logger.info("Shutting down the server")


def setup_cors(app):
    """Set up cors."""
    cors = aiohttp_cors.setup(
        app,
        defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True, expose_headers="*", allow_headers="*"
            )
        },
    )


clients = {}
plugins = {}


async def app_handler(request):
    return web.Response(text="clients: " + str(clients))


def setup_router(app, static_dir=None):
    if static_dir is not None:
        app.router.add_static("/", path=str(static_dir))
    app.router.add_get("/apps", app_handler)


finalizers = []


def setup_socketio(sio):
    @sio.event
    async def list_plugins(sid, data):
        return list(plugins.values())

    @sio.event
    async def register_plugin(sid, config):
        plugins[config["id"]] = config
        plugin_channel = str(uuid.uuid4())
        config["plugin_channel"] = plugin_channel
        clients = {}
        config["clients"] = clients

        # broadcast to the plugin message to all the clients
        @sio.on(plugin_channel)
        async def on_plugin_message(sid, data):
            if data.get("peer_id"):
                for k in clients:
                    if data.get("peer_id") == clients[k]["channel"]:
                        await sio.emit(clients[k]["channel"], data)
            else:
                for k in clients:
                    await sio.emit(clients[k]["channel"], data)

        async def finalize():
            if config["id"] in plugins:
                for k in clients:
                    await sio.emit(clients[k]["channel"], {"type": "disconnect"})
                del plugins[config["id"]]

        finalizers.append([sid, finalize])
        return {"channel": plugin_channel}

    @sio.event
    async def start_plugin(sid, config):
        for k in plugins:
            if plugins[k]["name"] == config["name"]:
                return connect_plugin(sid, {"id": plugins[k]["id"]})
        return {
            "error": "Plugin execution is disabled, you are only allowed to connect an existing plugin."
        }

    @sio.event
    async def connect_plugin(sid, data):
        pid = data.get("id")

        if pid in plugins:
            plugin_info = plugins[pid]
            client_info = {}

            # generate a channel and store it to plugin.clients
            client_channel = str(uuid.uuid4())
            logger.info(f"{client_channel}(sid:{sid}) is connecting to plugin {pid}")
            plugin_info["clients"][client_channel] = client_info
            client_info["channel"] = client_channel

            # listen to the client channel and forward to the plugin
            @sio.on(client_channel)
            async def on_client_message(sid, data):
                await sio.emit(plugin_info["plugin_channel"], data)

            # notify the plugin about the new client
            await sio.emit(plugin_info["plugin_channel"] + "-new-client", client_info)

            async def finalize():
                del plugin_info["clients"][client_channel]

            finalizers.append([sid, finalize])

            return {"channel": client_channel}
        else:
            logger.error(f"Plugin not found {pid}, requested by client {sid}")
            return {"error": "Plugin not found: " + pid}

    @sio.event
    async def connect(sid, environ):
        logger.info("%s connected", sid)

    @sio.event
    async def disconnect(sid):
        lst2finalize = [b for b in finalizers if b[0] == sid]
        for obj in lst2finalize:
            finalize_func = obj[1]
            try:
                logger.info("Removing " + obj[0])
                await finalize_func()
            finally:
                finalizers.remove(obj)


if __name__ == "__main__":
    app = create_socketio_server()
    web.run_app(app, port=9988)



"""Provide setup function to prepare the engine."""
import argparse
import os
from imjoy import __version__


def parse_cmd_line(args=None):
    """Parse the command line options."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--jupyter", action="store_true", help="start jupyter notebook server"
    )
    parser.add_argument("--token", type=str, default=None, help="connection token")
    parser.add_argument(
        "--random-token", action="store_true", help="randomly generate a token"
    )
    parser.add_argument("--debug", action="store_true", help="debug mode")
    parser.add_argument(
        "--serve",
        type=str,
        default=None,
        help="serve in production mode from the specified port",
    )
    parser.add_argument(
        "--plugin-dir", type=str, default=None, help="path to a plugin folder"
    )
    parser.add_argument(
        "--plugin-file", type=str, default=None, help="path to a plugin file"
    )
    parser.add_argument(
        "--plugin-server",
        type=str,
        default=None,
        help="url to the plugin socketio server",
    )

    parser.add_argument(
        "--host", type=str, default="127.0.0.1", help="jupyter server host"
    )
    parser.add_argument(
        "--base-url",
        type=str,
        default=None,
        help="the base url for accessing this plugin engine",
    )
    parser.add_argument(
        "--allow-origin",
        type=str,
        default="https://lib.imjoy.io",
        help="the allow origin header to prevent unintended access from other website",
    )
    parser.add_argument(
        "--content-security-policy",
        type=str,
        default="frame-ancestors 'self' https://imjoy.io https://*.imjoy.io",
        help="the Content-Security-Policy header to prevent unintended access from other website",
    )
    parser.add_argument("--port", type=str, default="9527", help="socketio port")
    parser.add_argument(
        "--workspace",
        type=str,
        default=os.path.abspath(os.getcwd()),
        help="workspace folder for plugins",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__
    )

    opt = parser.parse_args(args=args)

    if opt.base_url is None or opt.base_url == "":
        opt.base_url = "http://{}:{}".format(opt.host, opt.port)

    if opt.base_url.endswith("/"):
        opt.base_url = opt.base_url[:-1]

    return opt



"""Provide main entrypoint."""
import json
import os
import re
import subprocess
import sys
import asyncio
from aiohttp import web
import logging
import urllib.request
from imjoy_rpc import default_config

from imjoy.options import parse_cmd_line
from imjoy.utils import read_or_generate_token, write_token

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("main")
logger.setLevel(logging.INFO)


def load_plugin(plugin_file):
    """load plugin file"""
    try:
        import yaml
    except:
        logger.error(
            "It appears that your ImJoy installation is not complete, please reinstall it with 'pip install imjoy[socketio]'"
        )
        raise SystemExit
    if os.path.isfile(plugin_file):
        content = open(plugin_file).read()
    elif plugin_file.startswith("http"):
        with urllib.request.urlopen(plugin_file) as response:
            content = response.read().decode("utf-8")
        # remove query string
        plugin_file = plugin_file.split("?")[0]
    else:
        raise Exception("Invalid input plugin file path: {}".format(plugin_file))
    if plugin_file.endswith(".py"):
        filename, _ = os.path.splitext(os.path.basename(plugin_file))
        default_config["name"] = filename[:32]
        try:
            exec(content, globals())
            logger.info("Plugin executed")
        except Exception as e:
            logger.error("Failed to execute plugin %s", e)

    elif plugin_file.endswith(".imjoy.html"):
        # load config
        found = re.findall("<config (.*)>\n(.*)</config>", content, re.DOTALL)[0]
        if "json" in found[0]:
            plugin_config = json.loads(found[1])
        elif "yaml" in found[0]:
            plugin_config = yaml.safe_load(found[1])
        default_config.update(plugin_config)

        # load script
        found = re.findall("<script (.*)>\n(.*)</script>", content, re.DOTALL)[0]
        if "python" in found[0]:
            try:
                exec(found[1], globals())
                logger.info("Plugin executed")
            except Exception as e:
                logger.error("Failed to execute plugin %s", e)
        else:
            raise Exception(
                "Invalid script type ({}) in file {}".format(found[0], plugin_file)
            )
    else:
        raise Exception("Invalid script file type ({})".format(plugin_file))


def main():
    """Run main."""
    opt = parse_cmd_line()
    background_task = None

    if opt.plugin_file and (opt.plugin_server or opt.serve):

        async def start_plugin(app):
            default_config.update(
                {
                    "name": "ImJoy Plugin",
                    "plugin_server": opt.plugin_server
                    or "http://127.0.0.1:{}".format(opt.serve),
                }
            )

            load_plugin(opt.plugin_file)

        background_task = start_plugin

    if opt.serve:
        try:
            from imjoy.socketio_server import create_socketio_server
        except:
            logger.error(
                "It appears that your ImJoy installation is not complete, please reinstall it with 'pip install imjoy[socketio]'"
            )
            raise SystemExit
        if opt.plugin_server and not opt.plugin_server.endswith(opt.serve):
            print(
                "WARNING: the specified port ({}) does not match the one in the url ({})".format(
                    opt.serve, opt.plugin_server
                )
            )
        app = create_socketio_server()
        if background_task:
            app.on_startup.append(background_task)
        web.run_app(app, port=opt.serve)
    elif opt.plugin_file:
        loop = asyncio.get_event_loop()
        loop.create_task(background_task(app))
        loop.run_forever()
    elif opt.jupyter:
        sys.argv = sys.argv[:1]
        sys.argc = 1
        from notebook.notebookapp import NotebookApp

        kwargs = {
            "open_browser": False,
            "allow_origin": opt.allow_origin,
            "ip": opt.host,
            "notebook_dir": opt.workspace,
            "port": int(opt.port),
            "tornado_settings": {
                "headers": {
                    "Access-Control-Allow-Origin": opt.allow_origin,
                    "Content-Security-Policy": opt.content_security_policy,
                }
            },
        }

        if not opt.token:
            if not opt.random_token:
                opt.token = read_or_generate_token()
                kwargs["token"] = opt.token
        else:
            kwargs["token"] = opt.token

        app = NotebookApp.instance(**kwargs)
        app.initialize()
        if app.port != int(opt.port):
            print("\nWARNING: using a different port: {}.\n".format(app.port))
        write_token(app.token)
        app._token_generated = True
        app.start()

    elif not opt.legacy:
        print(
            "\nNote: We are migrating the backend of the ImJoy Engine to Jupyter, to use it please run `imjoy --jupyter`.\n\nIf you want to use the previous engine, run `imjoy --legacy`, however, please note that it maybe removed soon.\n"
        )


if __name__ == "__main__":
    main()
